package tech.mercantec.easyeat.helpers

import android.content.Context
import android.content.Intent
import android.util.Log
import android.widget.Toast
import kotlinx.serialization.Serializable
import kotlinx.serialization.SerializationException
import tech.mercantec.easyeat.BuildConfig
import java.net.HttpURLConnection
import java.net.URL
import kotlinx.serialization.json.*
import kotlinx.serialization.serializer
import tech.mercantec.easyeat.LoginActivity
import java.io.IOException

class ApiRequestException(message: String, cause: Throwable?) : Exception(message, cause)

class HttpResponse(val body: String, val code: Int)

fun request(ctx: Context, method: String, path: String, data: String?, autoRefresh: Boolean = true): HttpResponse {
    val url = URL(BuildConfig.API_BASE_URL + path)

    val prefs = ctx.getSharedPreferences("easyeat", Context.MODE_PRIVATE)
    val authToken = prefs.getString("auth-token", null)
    val refreshToken = prefs.getString("auth-token", null)

    try {
        with(url.openConnection() as HttpURLConnection) {
            requestMethod = method

            if (authToken != null)
                setRequestProperty("Authorization", "Bearer $authToken")

            if (data != null) {
                setRequestProperty("Content-Type", "application/json")

                outputStream.write(data.toByteArray())
                outputStream.flush()
            }

            if (responseCode == 401 && refreshToken != null) {
                if (!autoRefresh || !refreshAuthToken(ctx, refreshToken)) {
                    val intent = Intent(ctx, LoginActivity::class.java)
                    intent.flags = Intent.FLAG_ACTIVITY_NEW_TASK or Intent.FLAG_ACTIVITY_CLEAR_TASK
                    ctx.startActivity(intent)

                    throw ApiRequestException("Please sign in again", null)
                }

                return request(ctx, method, path, data, false)
            }

            if (responseCode >= 400)
                return HttpResponse(errorStream.readBytes().decodeToString(), responseCode)

            return HttpResponse(inputStream.readBytes().decodeToString(), responseCode)
        }
    } catch (e: IOException) {
        Log.e("EasyEat", e.toString())

        throw ApiRequestException("Failed to connect to server: " + e.message, e)
    }
}

@Serializable
class HttpErrorResponse(val message: String)

inline fun <reified Req, reified Res> requestJson(ctx: Context, method: String, path: String, data: Req?): Res {
    val requestJson =
        if (data != null)
            Json.encodeToString(serializer<Req>(), data)
        else null

    val response = request(ctx, method, path, requestJson)

    if (response.code >= 400) {
        try {
            val error = Json.decodeFromString<HttpErrorResponse>(response.body)

            throw ApiRequestException(error.message, null)
        } catch (e: SerializationException) {
            if (e.message != null)
                Log.e("EasyEat", e.message!!)
            Log.e("EasyEat", response.body)

            throw ApiRequestException("Request failed with HTTP status code ${response.code}", e)
        }
    }

    try {
        return Json.decodeFromString<Res>(response.body)
    } catch (e: SerializationException) {
        if (e.message != null)
            Log.e("EasyEat", e.message!!)
        Log.e("EasyEat", response.body)

        throw ApiRequestException("Failed to parse response: $response", e)
    }
}
